/*++
/* NAME
/*	smtpd_file 3
/* SUMMARY
/*	smtp server to file
/* SYNOPSIS
/*	#include "smtpd.h"
/*
/*	int	smtpd_file_init(state)
/*	SMTPD_STATE *state;
/* DESCRIPTION
/*	This module implements the smtp server to maildrop queue file
/*	interface.
/*
/*	smtpd_file_init() performs initializations and prepares for
/*	capturing a message to queue file.
/* BUGS
/*	Each mail program should be able to write a message to queue file.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>

/* Utility library. */

#include <msg.h>
#include <mymalloc.h>
#include <vstream.h>
#include <vstring.h>

/* Global library. */

#include <mail_queue.h>
#include <cleanup_user.h>
#include <mail_proto.h>
#include <opened.h>
#include <mail_error.h>
#include <mail_params.h>

/* Application-specific. */

#include "smtpd.h"

/* smtpd_file_open - open maildrop queue file */

static void smtpd_file_open(SMTPD_STATE *state)
{
    const char *path;
    const char *queue_id;

    /*
     * Open a maildrop file, and stash away a copy of the queue file pathname
     * so that we can clean up when the request is aborted.
     */
    state->cleanup =
	mail_queue_enter(MAIL_QUEUE_MAILDROP, 0600, &path, &queue_id);
    if (state->cleanup == 0)
	msg_fatal("create %s queue file: %m", MAIL_QUEUE_MAILDROP);
    smtpd_path = mystrdup(path);
    state->queue_id = mystrdup(queue_id);
}

/* smtpd_file_finish - finish maildrop file */

static void smtpd_file_finish(SMTPD_STATE *state)
{
    long    size = 0;			/* XXX stupid gcc */

    /*
     * Make sure the message makes it to file. Set the execute bit when no
     * write error was detected.
     */
    if (state->err == CLEANUP_STAT_OK
	&& ((size = vstream_ftell(state->cleanup)) < 0
#ifdef HAS_FSYNC
	    || fsync(vstream_fileno(state->cleanup))
#endif
	    || fchmod(vstream_fileno(state->cleanup), 0700)))
	state->err = CLEANUP_STAT_WRITE;

    /*
     * Close the queue file and mark it as closed. Be prepared for
     * vstream_fclose() to fail even after vstream_fflush() and fsync()
     * reported no error. Reason: after a file is closed, some networked file
     * systems copy the file out to another machine. Running the queue on a
     * remote file system is not recommended, if only for performance
     * reasons.
     */
    if (vstream_fclose(state->cleanup))
	state->err = CLEANUP_STAT_WRITE;
    state->cleanup = 0;

    /*
     * When all is well, prevent the file from being deleted upon cleanup,
     * and notify the pickup service that a new message has been queued. If
     * the mail system is down, the message will stay in the maildrop
     * directory for a while, so we syslog a message to that effect.
     */
    if (state->err == CLEANUP_STAT_OK) {
	myfree(smtpd_path);
	smtpd_path = 0;
	if (mail_trigger(MAIL_CLASS_PUBLIC, MAIL_SERVICE_PICKUP, "", 1) < 0)
	    opened(state->queue_id, state->sender, size, "queued");
    }
}

/* smtpd_file_init - register file access methods */

void    smtpd_file_init(SMTPD_STATE *state)
{
    state->open_message = smtpd_file_open;
    state->finish_message = smtpd_file_finish;
}
